from unittest import main, TestCase
from consts import *
from _core.json_convenience import *


class TestSuiteReadJSONFile(TestCase):

    def setUp(self):
        globalSetUp(write=True)

    def tearDown(self):
        globalTearDown()

    def test_raisesErrorIfPathInvalid(self):
        try:
            read_json_file(file_path=pathToNoJSON)
            self.fail("this should have raised a FileNotFoundError")
        except FileNotFoundError:
            pass
        except Exception:
            self.fail("this should have raised a FileNotFoundError")

    def test_raisesErrorIfJSONFileInvalid(self):
        try:
            read_json_file(file_path=pathToInvalidJSON)
            self.fail(msg="this should have raised a JSONDecodeError")
        except JSONDecodeError:
            pass
        except Exception:
            self.fail(msg="this should have raised a JSONDecodeError")

    def test_readDataEqualToFileContents(self):
        readData = read_json_file(file_path=pathToValidJSON)
        self.assertTrue(expr=readData == validJSONData, msg="read and decoded file contents are not correct")


class TestSuiteWriteJSONFile(TestCase):

    def setUp(self):
        globalSetUp(write=False)

    def tearDown(self):
        globalTearDown()

    def test_raisesErrorIfPathInvalid(self):
        try:
            write_json_file(file_path=pathToNoJSON, data=validJSONData)
            self.fail(msg="this should have raised a FileNotFoundError")
        except FileNotFoundError:
            pass
        except Exception:
            self.fail(msg="this should have raised a FileNotFoundError")

    def test_raisesErrorIfDataNotJSONSerializable(self):
        try:
            write_json_file(file_path=pathToValidJSON, data=invalidJSONData)
            self.fail(msg="this should have raised a NotAObjectError")
        except NotAObjectError:
            pass
        except Exception:
            self.fail(msg="this should have raised a NotAObjectError")

    def test_writtenDataEqualToInputData(self):
        write_json_file(file_path=pathToValidJSON, data=validJSONData)
        with open(file=pathToValidJSON, mode="r") as fp:
            self.assertTrue(expr=validJSONData == load(fp=fp), msg="written and encoded file contents are not correct")


if __name__ == "__main__":
    main()



from unittest import main, TestCase
from parameterized import parameterized
from consts import *
from _core.json_convenience import *


class SuperTestClass(TestCase):

    def setUp(self):
        globalSetUp(write=True)

    def tearDown(self):
        globalTearDown()


class TestSuiteIsFormatCorrect(SuperTestClass):

    def test_raisesErrorIfPathInvalid(self):
        try:
            is_format_correct(file_path=pathToNoJSON)
            self.fail(msg="this should have raised a FileNotFoundError")
        except FileNotFoundError:
            pass
        except Exception:
            self.fail(msg="this should have raised a FileNotFoundError")

    def test_trueForValidJSONFile(self):
        self.assertTrue(expr=is_format_correct(file_path=pathToValidJSON),
                        msg="isFormatCorrect says format is not correct even though it is")

    def test_falseForInvalidJSONFile(self):
        self.assertFalse(expr=is_format_correct(file_path=pathToInvalidJSON),
                         msg="isFormatCorrect says format is correct even though it is not")


class TestSuiteIndentJSONFile(SuperTestClass):

    def test_raisesErrorIfPathInvalid(self):
        try:
            indent_json_file(file_path=pathToNoJSON)
            self.fail(msg="this should have raised a FileNotFoundError")
        except FileNotFoundError:
            pass
        except Exception:
            self.fail(msg="this should have raised a FileNotFoundError")

    def test_indentsFileCorrectly(self):
        indent_json_file(file_path=pathToValidJSON)
        with open(file=pathToValidJSON, mode="r") as fp:
            fileStr = fp.read()
        self.assertTrue(expr=fileStr == indentedValidJSONStr, msg="the file is not correctly indented")


class TestSuiteGetProperty(SuperTestClass):

    def test_raisesErrorIfPathInvalid(self):
        try:
            get_property(file_path=pathToNoJSON, keys=())
            self.fail(msg="this should have raised a FileNotFoundError")
        except FileNotFoundError:
            pass
        except Exception:
            self.fail(msg="this should have raised a FileNotFoundError")

    def test_raisesErrorIfInvalidJSONInFile(self):
        try:
            get_property(file_path=pathToInvalidJSON, keys=())
            self.fail(msg="this should have raised a JSONDecodeError")
        except JSONDecodeError:
            pass
        except Exception:
            self.fail(msg="this should have raised a JSONDecodeError")

    def test_raisesErrorIfNotAProperty(self):
        try:
            get_property(file_path=pathToValidJSON, keys=(objectKey,))
            self.fail(msg="this should have raised a NotAPropertyError")
        except NotAPropertyError:
            pass
        except Exception:
            self.fail(msg="this should have raised a NotAPropertyError")

    def test_raisesErrorIfAKeyDoesNotExist(self):
        try:
            get_property(file_path=pathToValidJSON, keys=invalidKeys)
            self.fail(msg="this should have raised a JSONKeyNotFoundError")
        except JSONKeyNotFoundError:
            pass
        except Exception:
            self.fail(msg="this should have raised a JSONKeyNotFoundError")

    @parameterized.expand(simpleTypeKeys)
    def test_returnsSimpleJSONPropertiesCorrectly(self, key: str):
        self.assertTrue(expr=validJSONData[key] == get_property(file_path=pathToValidJSON, keys=(key,)),
                        msg="getProperty return wrong value")

    def test_returnsJSONArrayCorrectly(self):
        self.assertTrue(expr=validJSONData[arrayKey] == get_property(file_path=pathToValidJSON, keys=(arrayKey,)),
                        msg="getProperty return wrong value")


class TestSuiteSetProperty(SuperTestClass):

    def test_raisesErrorIfPathInvalid(self):
        try:
            set_property(file_path=pathToNoJSON, keys=(), value=None)
            self.fail(msg="this should have raised a FileNotFoundError")
        except FileNotFoundError:
            pass
        except Exception:
            self.fail(msg="this should have raised a FileNotFoundError")

    def test_raisesErrorIfInvalidJSONInFile(self):
        try:
            set_property(file_path=pathToInvalidJSON, keys=(), value=None)
            self.fail(msg="this should have raised a JSONDecodeError")
        except JSONDecodeError:
            pass
        except Exception:
            self.fail(msg="this should have raised a JSONDecodeError")

    def test_raisesErrorIfNotAProperty(self):
        try:
            set_property(file_path=pathToValidJSON, keys=(objectKey,), value=None)
            self.fail(msg="this should have raised a NotAPropertyError")
        except NotAPropertyError:
            pass
        except Exception:
            self.fail(msg="this should have raised a NotAPropertyError")

    def test_raisesErrorIfNewValueNotAProperty(self):
        try:
            set_property(file_path=pathToValidJSON, keys=(arrayKey,), value=nonEmptyDict)
            self.fail(msg="this should have raised a NotAPropertyError")
        except NotAPropertyError:
            pass
        except Exception:
            self.fail(msg="this should have raised a NotAPropertyError")

    def test_raisesErrorIfAKeyDoesNotExist(self):
        try:
            set_property(file_path=pathToValidJSON, keys=invalidKeys, value=None)
            self.fail(msg="this should have raised a JSONKeyNotFoundError")
        except JSONKeyNotFoundError:
            pass
        except Exception:
            self.fail(msg="this should have raised a JSONKeyNotFoundError")

    @parameterized.expand(list(simpleTypeKeysWithNewValues.keys()))
    def test_writesSimpleJSONPropertiesCorrectly(self, key: str):
        set_property(file_path=pathToValidJSON, keys=(key,), value=simpleTypeKeysWithNewValues[key])
        with open(file=pathToValidJSON, mode="r") as fp:
            self.assertTrue(expr=simpleTypeKeysWithNewValues[key] == load(fp=fp)[key],
                            msg="after calling setProperty value is not written correctly to the file")

    def test_writesJSONArrayCorrectly(self):
        set_property(file_path=pathToValidJSON, keys=(arrayKey,), value=newArray)
        with open(file=pathToValidJSON, mode="r") as fp:
            self.assertTrue(expr=newArray == load(fp=fp)[arrayKey],
                            msg="after calling setProperty value is not written correctly to the file")


class TestSuiteAddProperty(SuperTestClass):

    def test_raisesErrorIfPathInvalid(self):
        try:
            add_property(file_path=pathToNoJSON, keys=(), new_key="", value=None)
            self.fail(msg="this should have raised a FileNotFoundError")
        except FileNotFoundError:
            pass
        except Exception:
            self.fail(msg="this should have raised a FileNotFoundError")

    def test_raisesErrorIfInvalidJSONInFile(self):
        try:
            add_property(file_path=pathToInvalidJSON, keys=(), new_key="", value=None)
            self.fail(msg="this should have raised a JSONDecodeError")
        except JSONDecodeError:
            pass
        except Exception:
            self.fail(msg="this should have raised a JSONDecodeError")

    def test_raisesErrorIfNotAProperty(self):
        try:
            add_property(file_path=pathToValidJSON, keys=(), new_key="", value=invalidJSONData)
            self.fail(msg="this should have raised a NotAPropertyError")
        except NotAPropertyError:
            pass
        except Exception:
            self.fail(msg="this should have raised a NotAPropertyError")

    def test_raisesErrorIfParentObjectIsNoJSONObject(self):
        try:
            add_property(file_path=pathToValidJSON, keys=(arrayKey,), new_key="", value=newArray)
            self.fail(msg="this should have raised a NotAObjectError")
        except NotAObjectError:
            pass
        except Exception:
            self.fail(msg="this should have raised a NotAObjectError")

    def test_raisesErrorIfAKeyDoesNotExist(self):
        try:
            add_property(file_path=pathToValidJSON, keys=invalidKeys, new_key="", value=None)
            self.fail(msg="this should have raised a JSONKeyNotFoundError")
        except JSONKeyNotFoundError:
            pass
        except Exception:
            self.fail(msg="this should have raised a JSONKeyNotFoundError")

    def test_raisesErrorIfNewKeyAlreadyExists(self):
        try:
            add_property(file_path=pathToValidJSON, keys=(), new_key=arrayKey, value=None)
            self.fail(msg="this should have raised a JSONKeyAlreadyExists")
        except JSONKeyAlreadyExists:
            pass
        except Exception:
            self.fail(msg="this should have raised a JSONKeyAlreadyExists")

    @parameterized.expand(mappedPythonTypes)
    def test_writesSimpleJSONPropertiesCorrectly(self, newVal: str):
        newKey = "newKey"
        add_property(file_path=pathToValidJSON, keys=(), new_key=newKey, value=newVal)
        with open(file=pathToValidJSON, mode="r") as fp:
            self.assertTrue(expr=newVal == load(fp=fp)[newKey],
                            msg="after calling addProperty the property is not written correctly in the json file")

    def test_writesArrayCorrectly(self):
        newKey = "newKey"
        add_property(file_path=pathToValidJSON, keys=(), new_key=newKey, value=newArray)
        with open(file=pathToValidJSON, mode="r") as fp:
            self.assertTrue(expr=newArray == load(fp=fp)[newKey],
                            msg="after calling addProperty the property is not written correctly in the json file")


class TestSuiteContainsProperty(SuperTestClass):

    def test_raisesErrorIfPathInvalid(self):
        try:
            contains_property(file_path=pathToNoJSON, keys=())
            self.fail(msg="this should have raised a FileNotFoundError")
        except FileNotFoundError:
            pass
        except Exception:
            self.fail(msg="this should have raised a FileNotFoundError")

    def test_raisesErrorIfInvalidJSONInFile(self):
        try:
            contains_property(file_path=pathToInvalidJSON, keys=())
            self.fail(msg="this should have raised a JSONDecodeError")
        except JSONDecodeError:
            pass
        except Exception:
            self.fail(msg="this should have raised a JSONDecodeError")

    def test_falseIfKeysEmpty(self):
        self.assertFalse(expr=contains_property(file_path=pathToValidJSON, keys=()),
                         msg="containsProperty returned True for a empty key set")

    def test_falseIfPropertyDoesntExist(self):
        self.assertFalse(expr=contains_property(file_path=pathToValidJSON, keys=invalidKeys),
                         msg="containsProperty returned True for a property that doesn't exist")

    def test_falseIfKeysPointToObject(self):
        self.assertFalse(expr=contains_property(file_path=pathToValidJSON, keys=(objectKey,)),
                         msg="key set points to json object, but contains Property returned True")

    @parameterized.expand(simpleTypeKeys)
    def test_trueIfPropertyExists(self, key: str):
        self.assertTrue(expr=contains_property(file_path=pathToValidJSON, keys=(key,)),
                        msg=f"contains Property says that property '{(key,)}' doesn't exist but it does")


class TestSuiteGetObject(SuperTestClass):

    def test_raisesErrorIfPathInvalid(self):
        try:
            get_object(file_path=pathToNoJSON, keys=())
            self.fail(msg="this should have raised a FileNotFoundError")
        except FileNotFoundError:
            pass
        except Exception:
            self.fail(msg="this should have raised a FileNotFoundError")

    def test_raisesErrorIfInvalidJSONInFile(self):
        try:
            get_object(file_path=pathToInvalidJSON, keys=())
            self.fail(msg="this should have raised a JSONDecodeError")
        except JSONDecodeError:
            pass
        except Exception:
            self.fail(msg="this should have raised a JSONDecodeError")

    @parameterized.expand(simpleTypeKeys)
    def test_raisesErrorIfNotAObject(self, key: str):
        try:
            get_object(file_path=pathToValidJSON, keys=(key,))
            self.fail(msg="this should have raised a NotAObjectError")
        except NotAObjectError:
            pass
        except Exception:
            self.fail(msg="this should have raised a NotAObjectError")

    def test_raisesErrorIfAKeyDoesNotExist(self):
        try:
            get_object(file_path=pathToValidJSON, keys=invalidKeys)
            self.fail(msg="this should have raised a JSONKeyNotFoundError")
        except JSONKeyNotFoundError:
            pass
        except Exception:
            self.fail(msg="this should have raised a JSONKeyNotFoundError")

    def test_returnsJSONObjectCorrectly(self):
        self.assertTrue(expr=validJSONData[objectKey] == get_object(file_path=pathToValidJSON, keys=(objectKey,)),
                        msg="getObject return wrong value")


class TestSuiteSetObject(SuperTestClass):

    def test_raisesErrorIfPathInvalid(self):
        try:
            setObject(file_path=pathToNoJSON, keys=(), new_object={})
            self.fail(msg="this should have raised a FileNotFoundError")
        except FileNotFoundError:
            pass
        except Exception:
            self.fail(msg="this should have raised a FileNotFoundError")

    def test_raisesErrorIfInvalidJSONInFile(self):
        try:
            setObject(file_path=pathToInvalidJSON, keys=(), new_object={})
            self.fail(msg="this should have raised a JSONDecodeError")
        except JSONDecodeError:
            pass
        except Exception:
            self.fail(msg="this should have raised a JSONDecodeError")

    @parameterized.expand(simpleTypeKeys)
    def test_raisesErrorIfNotAObject(self, key: str):
        try:
            setObject(file_path=pathToValidJSON, keys=(key,), new_object={})
            self.fail(msg="this should have raised a NotAObjectError")
        except NotAObjectError:
            pass
        except Exception:
            self.fail(msg="this should have raised a NotAObjectError")

    def test_raisesErrorIfNewValueNotAObject(self):
        try:
            setObject(file_path=pathToValidJSON, keys=(objectKey,), new_object=newArray)
            self.fail(msg="this should have raised a NotAObjectError")
        except NotAObjectError:
            pass
        except Exception:
            self.fail(msg="this should have raised a NotAObjectError")

    def test_raisesErrorIfAKeyDoesNotExist(self):
        try:
            setObject(file_path=pathToValidJSON, keys=invalidKeys, new_object={})
            self.fail(msg="this should have raised a JSONKeyNotFoundError")
        except JSONKeyNotFoundError:
            pass
        except Exception:
            self.fail(msg="this should have raised a JSONKeyNotFoundError")

    def test_writesJSONObjectCorrectly(self):
        setObject(file_path=pathToValidJSON, keys=(objectKey,), new_object=nonEmptyDict)
        with open(file=pathToValidJSON, mode="r") as fp:
            self.assertTrue(expr=nonEmptyDict == load(fp=fp)[objectKey],
                            msg="after calling setObject object is not written correctly to the file")

    def test_writesEmptyJSONObjectCorrectly(self):
        setObject(file_path=pathToValidJSON, keys=(objectKey,), new_object=emptyDict)
        with open(file=pathToValidJSON, mode="r") as fp:
            self.assertTrue(expr=emptyDict == load(fp=fp)[objectKey],
                            msg="after calling setObject object is not written correctly to the file")


class TestSuiteAddObject(SuperTestClass):

    def test_raisesErrorIfPathInvalid(self):
        try:
            add_object(file_path=pathToNoJSON, keys=(), new_key="", new_object={})
            self.fail(msg="this should have raised a FileNotFoundError")
        except FileNotFoundError:
            pass
        except Exception:
            self.fail(msg="this should have raised a FileNotFoundError")

    def test_raisesErrorIfInvalidJSONInFile(self):
        try:
            add_object(file_path=pathToInvalidJSON, keys=(), new_key="", new_object={})
            self.fail(msg="this should have raised a JSONDecodeError")
        except JSONDecodeError:
            pass
        except Exception:
            self.fail(msg="this should have raised a JSONDecodeError")

    def test_raisesErrorIfParentObjectIsNoJSONObject(self):
        try:
            add_object(file_path=pathToValidJSON, keys=(arrayKey,), new_key="", new_object=nonEmptyDict)
            self.fail(msg="this should have raised a NotAObjectError")
        except NotAObjectError:
            pass
        except Exception:
            self.fail(msg="this should have raised a NotAObjectError")

    @parameterized.expand(mappedPythonTypes)
    def test_raisesErrorIfNotAObject(self, test_object: any):
        try:
            add_object(file_path=pathToValidJSON, keys=(), new_key="", new_object=test_object)
            self.fail(msg="this should have raised a NotAObjectError")
        except NotAObjectError:
            pass
        except Exception:
            self.fail(msg="this should have raised a NotAObjectError")

    def test_raisesErrorIfAKeyDoesNotExist(self):
        try:
            add_object(file_path=pathToValidJSON, keys=invalidKeys, new_key="", new_object={})
            self.fail(msg="this should have raised a JSONKeyNotFoundError")
        except JSONKeyNotFoundError:
            pass
        except Exception:
            self.fail(msg="this should have raised a JSONKeyNotFoundError")

    def test_raisesErrorIfNewKeyAlreadyExists(self):
        try:
            add_object(file_path=pathToValidJSON, keys=(), new_key=objectKey, new_object={})
            self.fail(msg="this should have raised a JSONKeyAlreadyExists")
        except JSONKeyAlreadyExists:
            pass
        except Exception:
            self.fail(msg="this should have raised a JSONKeyAlreadyExists")

    def test_writesJSONObjectCorrectly(self):
        newKey = "newKey"
        add_object(file_path=pathToValidJSON, keys=(), new_key=newKey, new_object=nonEmptyDict)
        with open(file=pathToValidJSON, mode="r") as fp:
            self.assertTrue(expr=nonEmptyDict == load(fp=fp)[newKey],
                            msg="after calling addObject the property is not written correctly in the json file")

    def test_writesEmptyJSONObjectCorrectly(self):
        newKey = "newKey"
        add_object(file_path=pathToValidJSON, keys=(), new_key=newKey, new_object=emptyDict)
        with open(file=pathToValidJSON, mode="r") as fp:
            self.assertTrue(expr=emptyDict == load(fp=fp)[newKey],
                            msg="after calling addObject the property is not written correctly in the json file")


class TestSuiteContainsObject(SuperTestClass):

    def test_raisesErrorIfPathInvalid(self):
        try:
            contains_object(file_path=pathToNoJSON, keys=())
            self.fail(msg="this should have raised a FileNotFoundError")
        except FileNotFoundError:
            pass
        except Exception:
            self.fail(msg="this should have raised a FileNotFoundError")

    def test_raisesErrorIfInvalidJSONInFile(self):
        try:
            contains_property(file_path=pathToInvalidJSON, keys=())
            self.fail(msg="this should have raised a JSONDecodeError")
        except JSONDecodeError:
            pass
        except Exception:
            self.fail(msg="this should have raised a JSONDecodeError")

    def test_trueIfKeysEmpty(self):
        self.assertTrue(expr=contains_object(file_path=pathToValidJSON, keys=()),
                        msg="containsObject returned False for a empty key set")

    def test_falseIfObjectDoesntExist(self):
        self.assertFalse(expr=contains_object(file_path=pathToValidJSON, keys=invalidKeys),
                         msg="containsObject returned True for a object that doesn't exist")

    @parameterized.expand(simpleTypeKeys)
    def test_falseIfKeysPointToProperty(self, key: str):
        self.assertFalse(expr=contains_object(file_path=pathToValidJSON, keys=(key,)),
                         msg="key set points to json property, but containsObject returned True")

    def test_trueIfObjectExists(self):
        self.assertTrue(expr=contains_object(file_path=pathToValidJSON, keys=(objectKey,)),
                        msg=f"containsObject says that object '{(objectKey,)}' doesn't exist but it does")


if __name__ == "__main__":
    main()



from unittest import main, TestCase
from parameterized import parameterized
from consts import *
import _core.json_convenience as jsonx


class TestSuiteIsJSONProperty(TestCase):

    @parameterized.expand(mappedPythonTypes)
    def test_trueForMappedPythonTypes(self, obj: any):
        self.assertTrue(expr=jsonx._is_json_property(raw_data=obj),
                        msg=f"the python object '{obj}' is a JSON property, but _isJSONProperty says otherwise")

    def test_trueForList(self):
        obj = nonEmptyList
        self.assertTrue(expr=jsonx._is_json_property(raw_data=obj),
                        msg=f"the python object '{obj}' is a JSON property, but _isJSONProperty says otherwise")

    def test_trueForEmptyList(self):
        obj = emptyList
        self.assertTrue(expr=jsonx._is_json_property(raw_data=obj),
                        msg=f"the python object '{obj}' is a JSON property, but _isJSONProperty says otherwise")

    def test_falseForDict(self):
        obj = nonEmptyDict
        self.assertFalse(expr=jsonx._is_json_property(raw_data=obj),
                         msg=f"the python object '{obj}' is no JSON property, but _isJSONProperty says otherwise")

    def test_falseForEmptyDict(self):
        obj = emptyDict
        self.assertFalse(expr=jsonx._is_json_property(raw_data=obj),
                         msg=f"the python object '{obj}' is no JSON property, but _isJSONProperty says otherwise")


class TestSuiteIsJSONObject(TestCase):

    @parameterized.expand(mappedPythonTypes)
    def test_falseForMappedPythonTypes(self, obj: any):
        self.assertFalse(expr=jsonx._is_json_object(raw_data=obj),
                         msg=f"the python object '{obj}' is no JSON object, but _isJSONObject says otherwise")

    def test_falseForList(self):
        obj = nonEmptyList
        self.assertFalse(expr=jsonx._is_json_object(raw_data=obj),
                         msg=f"the python object '{obj}' is no JSON object, but _isJSONObject says otherwise")

    def test_falseForEmptyList(self):
        obj = emptyList
        self.assertFalse(expr=jsonx._is_json_object(raw_data=obj),
                         msg=f"the python object '{obj}' is no JSON object, but _isJSONObject says otherwise")

    def test_trueForDict(self):
        obj = nonEmptyDict
        self.assertTrue(expr=jsonx._is_json_object(raw_data=obj),
                        msg=f"the python object '{obj}' is a JSON object, but _isJSONObject says otherwise")

    def test_trueForEmptyDict(self):
        obj = emptyDict
        self.assertTrue(expr=jsonx._is_json_object(raw_data=obj),
                        msg=f"the python object '{obj}' is a JSON object, but _isJSONObject says otherwise")


class TestSuiteContainsKey(TestCase):

    def test_trueForKeyInDict(self):
        key = list(nonEmptyDict.keys())[0]
        self.assertTrue(expr=jsonx._contains_key(raw_data=nonEmptyDict, key=key),
                        msg=f"The dict '{nonEmptyDict}' contains key '{key}', but _containsKey says otherwise")

    def test_falseForEmptyDict(self):
        key = "key"
        self.assertFalse(expr=jsonx._contains_key(raw_data=emptyDict, key=key),
                         msg=f"The dict '{nonEmptyDict}' is empty, but _containsKey says it contains key '{key}'")


class TestSuiteGetValueOfKeys(TestCase):

    def test_returnsObjectForEmptyKeysList(self):
        self.assertTrue(expr=validJSONData == jsonx._get_value_of_keys(raw_data=validJSONData, keys=()),
                        msg="_getValueOfKeys should return whole obj for empty keys set")

    @parameterized.expand(simpleTypeKeys)
    def test_findsAndReturnsSimpleType(self, key: str):
        self.assertTrue(expr=validJSONData[key] == jsonx._get_value_of_keys(raw_data=validJSONData, keys=(key,)),
                        msg=f"_getValueOfKeys does not return key '{key}' correctly")

    def test_findsAndReturnsList(self):
        self.assertTrue(
            expr=validJSONData[arrayKey] == jsonx._get_value_of_keys(raw_data=validJSONData, keys=(arrayKey,)),
            msg=f"_getValueOfKeys does not return key '{arrayKey}' correctly")

    def test_findsAndReturnsDict(self):
        self.assertTrue(
            expr=validJSONData[objectKey] == jsonx._get_value_of_keys(raw_data=validJSONData, keys=(objectKey,)),
            msg=f"_getValueOfKeys does not return key '{objectKey}' correctly")

    def test_findsAndReturnsNonTopLevelMember(self):
        self.assertTrue(expr=nonTopLevelValue == jsonx._get_value_of_keys(raw_data=validJSONData, keys=nonTopLevelKeys),
                        msg=f"_getValueOfKeys does not return keys '{nonTopLevelKeys}' correctly")

    def test_raisesErrorIfAKeyDoesNotExist(self):
        try:
            jsonx._get_value_of_keys(raw_data=validJSONData, keys=invalidKeys)
            self.fail(msg="this should have raised a JSONKeyNotFoundError")
        except jsonx.JSONKeyNotFoundError:
            pass
        except Exception:
            self.fail(msg="this should have raised a JSONKeyNotFoundError")


if __name__ == "__main__":
    main()



import setuptools

with open("README.md", "r") as fp:
    long_description = fp.read()

setuptools.setup(
    name="json-convenience",
    version="0.0.0",
    author="Nils Urbach",
    author_email="ndu01u@gmail.com",
    description="additional methods for handling json files in python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    keywords=[
        "json",
        "extended",
        "convenience"
    ],
    url="https://github.com/Schnilsibus/jsonExtended.git",
    package_dir={"": "_core"},
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Topic :: Text Processing",
        "Topic :: Utilities",
    ],
    test_suite="tests"
)



"""
Name
----
    json-convenience
Description
-----------
    This module extends the json module.
    Its core functionality is to get, set or add single json objects or properties from a json file,
    without manually loading and writing to the file and working with it's the dict representation.
    This could for example be used for configurations files o.ä.

Classes
-------
    no classes

Functions
-------
    read_json_file() - read the contents of a json file
    write_json_file() - write a dict to a json file
    is_format_correct() - check if a file contains valid json
    indent_json_file() - pretty indent a json file
    get_property() - get a specific property in a json file
    set_property() - set a specific property in a json file
    add_property() - add a specific property to a json file
    contains_property() - check if a json file contains a specific property
    get_object() - get a specific object in a json file
    set_object() - set a specific object in a json file
    add_object() - add a specific object to a json file
    contains_object() - check if a json file contains a specific object

Exceptions
----------
    NotAPropertyError - raised if a python/json object is not a (or not mapped to) a json property
    NotAObjectError - raised if a python/json object is not a (or not mapped to) a json object
    JSONKeyNotFoundError - raised if a json key is not found a file
    JSONKeyAlreadyExistsError - raised if a json key already exists & it is tried to add it to a file
"""

from json import dump, load, JSONDecodeError
from pathlib import Path
from typing import Any, Union, Tuple

_indent_level = 4

Property = Union[None, bool, int, float, str, list]
Object = dict


class NotAPropertyError(Exception):
    """
    An Error that is raised if a python object has a type that is not mapped to a json type

    Can also be raised if contents of a json file should be a json property but are a json object.
    json types that are called properties are: array, string, number, boolean, null.
    The python types that are mapped to those are: list, str, int, float, bool, None.
    Every other python type is not considered a json property.
    """

    def __init__(self, no_property_object: Any):
        """
        Parameters
        ----------
        no_property_object: Any
            the python object whose type is not mapped to a json property
        """

        Exception.__init__(
            self,
            f"the json object {no_property_object} is not a property; properties are: json data types and json arrays"
        )


class NotAObjectError(Exception):
    """
    An Error that is raised if a python object has a type that is not mapped to a json object

    Can also be raised if contents of a json file should be a json object but are a json property.
    json objects are the objects enclosed in curly braces in json files.
    The python type that is mapped to this is: dict.
    Every other python type is not considered a json object.
    """

    def __init__(self, no_object: Any):
        """
        Parameters
        ----------
        no_object: Any
            the python object whose type is not mapped to a json object
        """

        Exception.__init__(self, f"the json value {no_object} is not a json object")


class JSONKeyNotFoundError(Exception):
    """
    An Error that is raised if a key is not found in a json file
    """

    def __init__(self, wrong_key: str, all_keys_of_object: Tuple, found_keys: Tuple = None):
        """
        Parameters
        ----------
        wrong_key: str
            the key that could not be found
        all_keys_of_object: Tuple
            all the keys of the object in which the wrong key could not be found
        found_keys: Tuple (default=None)
            all the keys of the parent json objects of the object in which the wrong key could not be found
        """

        Exception.__init__(self,
                           f"key '{wrong_key}' not in {all_keys_of_object}; found keys: [{'->'.join(found_keys)}]")


class JSONKeyAlreadyExists(Exception):
    """
    An Error that is raised if a key already exists in a json object
    """

    def __init__(self, double_key: str, all_keys_of_object: Tuple, found_keys: Tuple = None):
        """
        Parameters
        ----------
        double_key: str
            the key that already exists
        all_keys_of_object: Tuple
            all the keys of the object that already contains the doubled key
        found_keys: Tuple
            all the keys of the parent json objects of the object that contains the doubled key
        """

        Exception.__init__(
            self,
            f"key '{double_key}' already exists in {all_keys_of_object}; found keys: [{'->'.join(found_keys)}]"
        )


def get_property(file_path: Path, keys: Tuple) -> Property:
    """
    Return a property in a json file

    Parameters
    ----------
    file_path: pathlib.Path
        the path to the json file
    keys: Tuple
        the ordered! keys in the json file that contain the desired property;
        the key of the property is the last element of the tuple

    Returns
    -------
    the specified property

    Raises
    ------
    FileNotFoundError
        if filePath doesn't point to a file
    json.JSONDecodeError
        if the file that filePath points to cannot be decoded by the json package ergo doesn't contain valid json
    JSONKeyNotFoundError
        if keys contains a key that cannot be found (in the json object that the keys before the not-found-key point to)
    NotAPropertyError
        if keys point to a json object (not a json property)
    """

    raw_data = read_json_file(file_path=file_path)
    value = _get_value_of_keys(raw_data=raw_data, keys=keys)
    if _is_json_object(raw_data=value):
        raise NotAPropertyError(no_property_object=value)
    return value


def set_property(file_path: Path, keys: Tuple, value: Property) -> None:
    """
    Set a property in a json file to a value

    Parameters
    ----------
    file_path: pathlib.Path
        the path to the json file
    keys: Tuple
        the ordered! keys in the json file that contain the desired property;
        the key of the property is the last element of the tuple
    value: Property (must be a python type that is mapped to json property)
        the new value of the property

    Returns
    -------
    None

    Raises
    ------
    FileNotFoundError
        if filePath doesn't point to a file
    json.JSONDecodeError
        if the file that filePath points to cannot be decoded by the json package ergo doesn't contain valid json
    JSONKeyNotFoundError
        if keys contains a key that cannot be found (in the json object that the keys before the not-found-key point to)
    NotAPropertyError
        - if keys point to a json object (not a json property)
        - if type(value) is not mapped to a json property
    """

    raw_data = read_json_file(file_path=file_path)
    parent_object = _get_value_of_keys(raw_data=raw_data, keys=keys[:-1])
    if not _contains_key(raw_data=parent_object, key=keys[-1]):
        raise JSONKeyNotFoundError(wrong_key=keys[-1], all_keys_of_object=tuple(parent_object.keys()),
                                   found_keys=keys[:-1])
    elif _is_json_object(raw_data=parent_object[keys[-1]]):
        raise NotAPropertyError(no_property_object=parent_object[keys[-1]])
    elif not _is_json_property(raw_data=value):
        raise NotAPropertyError(no_property_object=value)
    parent_object[keys[-1]] = value
    write_json_file(file_path=file_path, data=raw_data)


def add_property(file_path: Path, keys: Tuple, new_key: str, value: Property) -> None:
    """
    Add a json property to a json file

    Parameters
    ----------
    file_path: pathlib.Path
        the path to the json file
    keys: Tuple
        the ordered! keys in the json file point to the json object that should contain the new property
    new_key: str
        the new key of the property
    value: Property (must be a python type that is mapped to json property)
        the value of the new property

    Returns
    -------
    None

    Raises
    ------
    FileNotFoundError
        if filePath doesn't point to a file
    json.JSONDecodeError
        if the file that filePath points to cannot be decoded by the json package ergo doesn't contain valid json
    JSONKeyNotFoundError
        if keys contains a key that cannot be found (in the json object that the keys before the not-found-key point to)
    NotAObjectError
        if keys points to a json property instead of a json object (can't add a property to a property)
    NotAPropertyError
        if the type(value) is not mapped to a json property
    """

    raw_data = read_json_file(file_path=file_path)
    parent_object = _get_value_of_keys(raw_data=raw_data, keys=keys)
    if not _is_json_object(raw_data=parent_object):
        raise NotAObjectError(no_object=parent_object)
    elif _contains_key(raw_data=parent_object, key=new_key):
        raise JSONKeyAlreadyExists(double_key=new_key, all_keys_of_object=tuple(parent_object.keys()), found_keys=keys)
    elif not _is_json_property(raw_data=value):
        raise NotAPropertyError(no_property_object=value, )
    parent_object[new_key] = value
    write_json_file(file_path=file_path, data=raw_data)


def contains_property(file_path: Path, keys: Tuple) -> bool:
    """
    Check if a json file contains a specified property

    Parameters
    ----------
    file_path: pathlib.Path
        the path to the json file
    keys: Tuple
        the ordered! keys in the json file that contain the desired property;
        the key of the property is the last element of the tuple
    
    Returns
    -------
    - True if the json file contains the specified property
    - False else

    Raises
    ------
    FileNotFoundError
        if filePath doesn't point to a file
    json.JSONDecodeError
        if the file that filePath points to cannot be decoded by the json package ergo doesn't contain valid json
    """

    raw_data = read_json_file(file_path=file_path)
    try:
        value = _get_value_of_keys(raw_data=raw_data, keys=keys)
        return _is_json_property(raw_data=value)
    except JSONKeyNotFoundError:
        return False


def get_object(file_path: Path, keys: Tuple) -> Object:
    """
    Return an object in a json file

    Parameters
    ----------
    file_path: pathlib.Path
        the path to the json file
    keys: Tuple
        the ordered! keys in the json file that contain the desired object;
        the key of the object is the last element of the tuple

    Returns
    -------
    the specified object

    Raises
    ------
    FileNotFoundError
        if filePath doesn't point to a file
    json.JSONDecodeError
        if the file that filePath points to cannot be decoded by the json package ergo doesn't contain valid json
    JSONKeyNotFoundError
        if keys contains a key that cannot be found (in the json object that the keys before the not-found-key point to)
    NotAObjectError
        if keys point to a json property (not a json object)
    """

    rawData = read_json_file(file_path=file_path)
    json_object = _get_value_of_keys(raw_data=rawData, keys=keys)
    if _is_json_property(raw_data=json_object):
        raise NotAObjectError(no_object=json_object)
    return json_object


def setObject(file_path: Path, keys: Tuple, new_object: Object) -> None:
    """
    Set an object in a json file

    Parameters
    ----------
    file_path: pathlib.Path
        the path to the json file
    keys: Tuple
        the ordered! keys in the json file that contain the desired object;
        the key of the object is the last element of the tuple
    new_object: dict
        the new value of the object

    Returns
    -------
    None

    Raises
    ------
    FileNotFoundError
        if filePath doesn't point to a file
    json.JSONDecodeError
        if the file that filePath points to cannot be decoded by the json package ergo doesn't contain valid json
    JSONKeyNotFoundError
        if keys contains a key that cannot be found (in the json object that the keys before the not-found-key point to)
    NotAObjectError
        - if keys point to a json property (not a json object)
        - if not type(value) == dict 
    """

    rawData = read_json_file(file_path=file_path)
    parentObject = _get_value_of_keys(raw_data=rawData, keys=keys[:-1])
    if not _contains_key(raw_data=parentObject, key=keys[-1]):
        raise JSONKeyNotFoundError(wrong_key=keys[-1], all_keys_of_object=tuple(parentObject.keys()),
                                   found_keys=keys[:-1])
    elif _is_json_property(raw_data=parentObject[keys[-1]]):
        raise NotAObjectError(no_object=parentObject[keys[-1]])
    elif not _is_json_object(raw_data=new_object):
        raise NotAObjectError(no_object=new_object)
    parentObject[keys[-1]] = new_object
    write_json_file(file_path=file_path, data=rawData)


def add_object(file_path: Path, keys: Tuple, new_key: str, new_object: Object) -> None:
    """
    Add a json object to a json file

    Parameters
    ----------
    file_path: pathlib.Path
        the path to the json file
    keys: Tuple
        the ordered! keys in the json file point to the json object that should contain the new object
    new_key: str
        the new key of the object
    new_object: dict
        the new object

    Returns
    -------
    None

    Raises
    ------
    FileNotFoundError
        if filePath doesn't point to a file
    json.JSONDecodeError
        if the file that filePath points to cannot be decoded by the json package ergo doesn't contain valid json
    JSONKeyNotFoundError
        if keys contains a key that cannot be found (in the json object that the keys before the not-found-key point to)
    NotAObjectError
        - if keys points to a json property instead of a json object (can't add an object to a property)
        - if the type(value) is not mapped to a json object
    """

    rawData = read_json_file(file_path=file_path)
    parentObject = _get_value_of_keys(raw_data=rawData, keys=keys)
    if not _is_json_object(raw_data=parentObject):
        raise NotAObjectError(no_object=parentObject)
    elif _contains_key(raw_data=parentObject, key=new_key):
        raise JSONKeyAlreadyExists(double_key=new_key, all_keys_of_object=tuple(parentObject.keys()), found_keys=keys)
    elif not _is_json_object(raw_data=new_object):
        raise NotAObjectError(no_object=new_object)
    parentObject[new_key] = new_object
    write_json_file(file_path=file_path, data=rawData)


def contains_object(file_path: Path, keys: Tuple) -> bool:
    """
    Check if a json file contains a specified object

    Parameters
    ----------
    file_path: pathlib.Path
        the path to the json file
    keys: Tuple
        the ordered! keys in the json file that contain the desired object;
        the key of the object is the last element of the tuple
    
    Returns
    -------
    - True if the json file contains the specified object
    - False else

    Raises
    ------
    FileNotFoundError
        if filePath doesn't point to a file
    json.JSONDecodeError
        if the file that filePath points to cannot be decoded by the json package ergo doesn't contain valid json
    """

    rawData = read_json_file(file_path=file_path)
    try:
        value = _get_value_of_keys(raw_data=rawData, keys=keys)
        return _is_json_object(raw_data=value)
    except JSONKeyNotFoundError:
        return False


def is_format_correct(file_path: Path) -> bool:
    """
    Check if a file contains valid json

    Parameters
    ----------
    file_path: pathlib.Path
        the path to the file

    Returns
    -------
    - True if the file contains valid json (== json.load() doesn't throw a json.JSONDecodeError)
    - False else

    Raises
    ------
    FileNotFoundError
        if filePath doesn't point to a file
    """

    try:
        read_json_file(file_path=file_path)
    except JSONDecodeError:
        return False
    return True


def indent_json_file(file_path: Path) -> None:
    f"""
    Indent a json file

    The file is indented using the json.dump() method with indent = {_indent_level}
    Parameters
    ----------
    file_path: pathlib.Path
        the path to the json file
    
    Returns
    -------
    None

    Raises
    ------
    FileNotFoundError
        if filePath doesn't point to a file
    json.JSONDecodeError
        if the file that filePath points to cannot be decoded by the json package ergo doesn't contain valid json
    """

    write_json_file(file_path=file_path, data=read_json_file(file_path=file_path))


def read_json_file(file_path: Path) -> Object:
    """
    Read the contents of a json file and returns them as a dict

    Parameters
    ----------
    file_path: pathlib.Path
        the path to the json file

    Returns
    -------
    The deserialized contents of a json file as a dict

    Raises
    ------
    FileNotFoundError
        if filePath doesn't point to a file
    json.JSONDecodeError
        if the file that filePath points to cannot be decoded by the json package ergo doesn't contain valid json 
    """

    if not file_path.exists():
        raise FileNotFoundError(f"the json file {file_path} doesn't exist")
    with file_path.open(mode="r") as fp:
        return load(fp=fp)


def write_json_file(file_path: Path, data: Object) -> None:
    """
    Write a dict to a json file

    Parameters
    ----------
    file_path: pathlib.Path
        the path to the json file
    data: dict
        the data that should be written to the json file
    Returns
    -------
    None

    Raises
    ------
    FileNotFoundError
        if filePath doesn't point to a file
    NotAObjectError
        if not type(data) == dict 
    """

    if not file_path.exists():
        raise FileNotFoundError(f"the json file {file_path} doesn't exist")
    elif not _is_json_object(raw_data=data):
        raise NotAObjectError(no_object=data)
    with file_path.open(mode="w") as fp:
        dump(obj=data, fp=fp, indent=_indent_level)


def _get_value_of_keys(raw_data: dict, keys: Tuple) -> Union[Object, Property]:
    current_object = raw_data
    for i in range(len(keys)):
        if not _contains_key(raw_data=current_object, key=keys[i]):
            raise JSONKeyNotFoundError(wrong_key=keys[i],
                                       all_keys_of_object=tuple(raw_data.keys()),
                                       found_keys=keys[:i])
        current_object = current_object[keys[i]]
    return current_object


def _is_json_property(raw_data: Any) -> bool:
    return type(raw_data) in [type(None), float, int, list, bool, str]


def _is_json_object(raw_data: Any) -> bool:
    return type(raw_data) == dict


def _contains_key(raw_data: Object, key: str) -> bool:
    return key in raw_data.keys()
